@@ -2,7 +2,7 @@
 Utilities module
 """
 from enum import Enum
-from typing import Dict
+from typing import Any, Dict, Iterable, List, Optional, Tuple
 import pandas as pd
 from python_graphql_client import GraphqlClient
 import numpy as np
@@ -14,7 +14,7 @@ import boto3
 from boto3.session import Session
 import os
 import string
-from os import path
+from os import DirEntry, path
 import requests
 import time
 import json
@@ -64,7 +64,7 @@ def set_api_key(key, key_file_path):
         json.dump(keys_json, f)
 
 
-def get_api_headers(canalyst_api_key: str) -> Dict[str, str]:
+def get_api_headers(canalyst_api_key: Optional[str]) -> Dict[str, str]:
     """
     Return the authorization bearer header to use for API requests and user agent
     """
@@ -89,7 +89,10 @@ class LogFile:
         self.log_file_name = f"{default_dir}/candas_logfile.csv"
 
         if not os.path.isfile(self.log_file_name):
-            rows = [["timestamp", "action"], [tm, "initiate logfile"]]
+            rows: Iterable[Iterable[Any]] = [
+                ["timestamp", "action"],
+                [tm, "initiate logfile"],
+            ]
             with open(self.log_file_name, "w", newline="") as csvfile:
                 csvwriter = csv.writer(csvfile)
                 csvwriter.writerows(rows)
@@ -114,7 +117,7 @@ class Getter:
     S3 Client to retreive data from S3
     """
 
-    def __init__(self, config: Config):
+    def __init__(self, config: Config) -> None:
         self.config = config
         self.log = LogFile(default_dir=self.config.default_dir)
         self.s3_client = boto3.client(
@@ -348,8 +351,8 @@ def filter_dataset(
         )
 
     if mrq is True:
-        mrq = df["MRQ"].iloc[0]
-        df = df.loc[df["period_name"] == df["MRQ"]]
+        mrq = df["MRFQ"].iloc[0]
+        df = df.loc[df["period_name"] == df["MRFQ"]]
 
     df = df.sort_values(["period_end_date", "name_index"])
 
@@ -377,14 +380,17 @@ def filter_dataset(
         list_numbers = []
         list_numbers = list(n_periods)
         df = df.groupby(["ticker", "time_series_name"]).take(list_numbers).reset_index()
+
     elif type(n_periods) is int:
         if n_periods > 0:
             df = (
-                df.groupby(["ticker", "time_series_name"]).tail(n_periods).reset_index()
+                df.groupby(["ticker", "time_series_name", "period_duration_type"])
+                .tail(n_periods)
+                .reset_index()
             )
         else:
             df = (
-                df.groupby(["ticker", "time_series_name"])
+                df.groupby(["ticker", "time_series_name", "period_duration_type"])
                 .head(-1 * n_periods)
                 .reset_index()
             )
@@ -394,12 +400,9 @@ def filter_dataset(
         + df["period_name"].str.split("-").str[0]
     )
 
-    if time_series_name == "":
-        if pivot == True:
-            return pivot_df(df, mrq_notation)
-        return df
-
-    if type(time_series_name) is not list and time_series_name != "":
+    if (
+        type(time_series_name) is not list and time_series_name != ""
+    ):  # this should be replaced with a search function
         df1 = df[
             df["time_series_name"].str.contains(
                 time_series_name, flags=re.IGNORECASE, regex=True
@@ -416,14 +419,6 @@ def filter_dataset(
 
         df = df.sort_values("period_end_date")
 
-        if n_diff != "":
-            df = df.sort_values(["ticker", "period_end_date"])
-            df["value"] = df["value"].pct_change(periods=int(n_diff))
-
-        if pivot is True:
-            df = pivot_df(df, mrq_notation)
-        return df
-
     if type(time_series_name) is list:
         list_df = []
         for item in time_series_name:
@@ -432,9 +427,11 @@ def filter_dataset(
             list_df.append(df_copy)
         df = pd.concat(list_df)
 
-        if n_diff != "":
-            df = df.sort_values(["ticker", "period_end_date"])
-            df["value"] = df["value"].pct_change(periods=int(n_diff))
+    if (
+        n_diff != ""
+    ):  # this should be replaced with a diff function in the new ModelFrame() class
+        df = df.sort_values(["ticker", "time_series_name", "period_end_date"])
+        df["value"] = df["value"].pct_change(periods=int(n_diff))
 
     if pivot is True:
         df = pivot_df(df, mrq_notation)
@@ -495,10 +492,7 @@ def map_scenario_urls(json):
     return scenario_map
 
 
-def get_name_index_from_csv(
-    ticker,
-    config,
-):
+def get_name_index_from_csv(ticker, config):
     default_dir = config.default_dir
 
     # check if the model exists if so use it.
@@ -589,8 +583,8 @@ def get_company_info_from_ticker(ticker, auth_headers, mds_host=settings.MDS_HOS
     response = requests.get(company_url, headers=auth_headers, verify=False)
 
     json = response.json().get("results")[0]
-    csin = json.get("csin")
-    latest_version = (
+    csin: str = json.get("csin")
+    latest_version: str = (
         json.get("latest_equity_model", {}).get("model_version", {}).get("name", {})
     )
 
@@ -925,9 +919,8 @@ def check_throttle_limit(response):
     return False
 
 
-def scandir(dir, ext):  # dir: str, ext: list
+def scandir(dir: str, ext: list) -> Tuple[List[str], List[str]]:  # dir: str, ext: list
     subfolders, files = [], []
-
     for f in os.scandir(dir):
         if f.is_dir():
             subfolders.append(f.path)
@@ -937,9 +930,9 @@ def scandir(dir, ext):  # dir: str, ext: list
 
     # to iterate is human, to recurse is divine
     for dir in list(subfolders):
-        sf, f = scandir(dir, ext)
+        sf, f = scandir(dir, ext)  # type: ignore
         subfolders.extend(sf)
-        files.extend(f)
+        files.extend(f)  # type: ignore
     return subfolders, files
 
 
@@ -1002,7 +995,6 @@ def create_drivers_dot(time_series, auth_headers, ticker, config=None, s3_client
         r = requests.get(url, headers=auth_headers, verify=False)
         open(path_name, "wb").write(r.content)
         return path_name
-    return
 
 
 def get_model_url(ticker, wp_host=settings.WP_HOST):
@@ -1168,50 +1160,132 @@ def json_to_df(dict_json, ticker):
 
 
 def mrq_df(df):
-    df = df[df["is_historical"] == True]
-
-    n_size = df.groupby(["ticker", "time_series_name"]).size().mean()
-    df["n"] = df.groupby(["ticker", "time_series_name"]).cumcount() - (n_size - 1)
-    df["n"] = df["n"].astype(int)
-
-    cols = list(df.groupby("n").first().reset_index()["n"])
-
-    df = pd.pivot_table(
-        df,
-        values="value",
-        index=[
-            "name_index",
-            # "model_name",
-            "ticker",
-            "category",
-            "time_series_name",
-            "time_series_description",
-            "is_driver",
-            "MRQ",
-        ],
-        columns=["n"],
-        aggfunc=np.sum,
-    ).reset_index()
-    list_cols = []
-
+    # need to loop this by ticker
+    ticker_list = list(set(list(df["ticker"])))
+    df_out = []
+    df2 = df
     name_cols = [
         "name_index",
-        # "model_name",
         "ticker",
         "category",
         "time_series_name",
         "time_series_description",
         "is_driver",
-        "MRQ",
+        "MRFQ",
     ]
 
-    col_names = []
-    for col in df.columns:
-        if col not in name_cols:
-            col = "MRQ" + str(col)
-            col_names.append(col)
-    df.columns = name_cols + col_names
-    return df.sort_values("name_index")
+    for ticker in ticker_list:
+        df = df2.loc[df2["ticker"] == ticker]
+        mrq = df.iloc[0]["MRFQ"]
+        mry = mrq[-4:]
+        if int(mrq[-6]) != 4:
+            mry = str(int(mry) - 1)
+
+        df_copy = df.copy()
+        df = df.loc[df["period_duration_type"] == "fiscal_quarter"]
+
+        i_switch = 0  # handle that when we get here, we may only have fiscal years or only fiscal quarters or both
+
+        if df.shape[0] != 0:
+            i_switch = 1  # we have fiscal quarters
+
+            df = df.sort_values(["ticker", "time_series_name", "period_end_date"])
+            df["temp_n"] = np.where(df["period_name"] == mrq, 0, 1)
+            df["temp_c"] = df.groupby(["ticker", "time_series_name"]).cumcount() + 1
+
+            df = df.sort_values("temp_n")
+            if df[df["period_name"].str.contains(mrq)].shape[0] == 0:
+                i_value = 0
+            else:
+                i_value = df.iloc[0]["temp_c"]
+
+            df["n_index"] = df["temp_c"] - i_value
+            df["FYFQ"] = df["n_index"]
+            df = df.sort_values("n_index")
+            df = df.drop(columns=["temp_c", "n_index", "temp_n"])
+            df = pd.pivot_table(
+                df,
+                values="value",
+                index=[
+                    "name_index",
+                    "ticker",
+                    "category",
+                    "time_series_name",
+                    "time_series_description",
+                    "is_driver",
+                    "MRFQ",
+                ],
+                columns=["FYFQ"],
+                aggfunc=np.sum,
+            ).reset_index()
+
+            rename_cols = []
+            for col in df.columns:
+                if col not in name_cols:
+                    col = "FQ" + str(col)
+                    rename_cols.append(col)
+            df.columns = name_cols + rename_cols
+            df_q = df
+
+        df = df2.loc[df2["ticker"] == ticker]
+        df = df.loc[df["period_duration_type"] == "fiscal_year"]
+
+        if df.shape[0] != 0:
+            if i_switch:
+                i_switch = (
+                    2  # we have both fiscal quarters and years otherwise i switch is 0
+                )
+
+            df = df.sort_values(["ticker", "time_series_name", "period_end_date"])
+            df["temp_n"] = np.where(df["period_name"].str.contains(mry), 0, 1)
+            df["temp_c"] = df.groupby(["ticker", "time_series_name"]).cumcount() + 1
+            df = df.sort_values("temp_n")
+
+            if df[df["period_name"].str.contains(mry)].shape[0] == 0:
+                i_value = 0
+            else:
+                i_value = df.iloc[0]["temp_c"]
+
+            df["n_index"] = df["temp_c"] - i_value
+            df["FYFQ"] = df["n_index"]
+            df = df.sort_values("n_index")
+            df = df.drop(columns=["temp_c", "n_index", "temp_n"])
+            df = pd.pivot_table(
+                df,
+                values="value",
+                index=[
+                    "name_index",
+                    "ticker",
+                    "category",
+                    "time_series_name",
+                    "time_series_description",
+                    "is_driver",
+                    "MRFQ",
+                ],
+                columns=["FYFQ"],
+                aggfunc=np.sum,
+            ).reset_index()
+
+            rename_cols = []
+            for col in df.columns:
+                if col not in name_cols:
+                    col = "FY" + str(col)
+                    rename_cols.append(col)
+            df.columns = name_cols + rename_cols
+            df_y = df
+
+        if i_switch == 2:  # both
+            df = pd.merge(
+                df_q, df_y, how="inner", left_on=name_cols, right_on=name_cols
+            )
+        if i_switch == 1:  # quarters only
+            df = df_q
+        if i_switch == 0:  # years only
+            df = df_y
+        df_out.append(df)
+
+    df = pd.concat(df_out, join="inner", ignore_index=True)
+    return df.sort_values(["ticker", "time_series_name"])
 
 
 def pivot_df(df, mrq_notation):
@@ -1228,13 +1302,13 @@ def pivot_df(df, mrq_notation):
         df,
         values="value",
         index=[
-            # "model_name",
             "ticker",
             "name_index",
             "category",
             "time_series_name",
             "time_series_description",
             "is_driver",
+            "MRFQ",
         ],
         columns=["period_name_sorted"],
         aggfunc=np.sum,
@@ -1244,7 +1318,9 @@ def pivot_df(df, mrq_notation):
     return df
 
 
-def refresh_cache(config: Config = None, log: LogFile = None, s3_client: Getter = None):
+def refresh_cache(
+    config: Config = None, log: LogFile = None, s3_client: Getter = None
+) -> None:
     """
     Refresh the cache by deleting and pulling files from S3
     """
@@ -1388,7 +1464,7 @@ def get_drivers_from_api(mds_host, csin, model_version, api_headers):
 
     driver_url = "api/equity-model-series/{csin}/equity-models/{model_version}/forecast-periods/{period}/data-points/?is_driver=true&page_size=200"
     driver_endpoint = f"{mds_host}/{driver_url.format(csin=quote_plus(csin), model_version=quote_plus(model_version), period=quote_plus(forecast_period_name))}"
-    next_url = driver_endpoint
+    next_url: Optional[str] = driver_endpoint
 
     while next_url is not None:
         drivers, next_url = _get_drivers_from_next_url(next_url, api_headers, logger)
